import { __awaiter, __decorate, __param } from "tslib";
import { observable, volatile } from '@microsoft/fast-element';
import { DI } from '@microsoft/fast-foundation';
import { classNames } from '@microsoft/fast-web-utilities';
import { AbstractStore, AbstractStoreRoot, } from '../store';
import { ReportService, SubService } from './services';
/**
 * @internal
 */
let DefaultSubStore = class DefaultSubStore extends AbstractStore {
    constructor(service) {
        super();
        this.service = service;
        this.x = '';
        this.y = 0;
        this.z = 0;
        this.onXChanged = this.createListener('x-changed', (detail) => (this.commit.x = detail));
        this.onYChanged = this.createListener('y-changed', (detail) => (this.commit.y = detail));
        this.onZChanged = this.createListener('z-changed', (detail) => (this.commit.z = detail));
        this.onSelectSubEntity = this.createListener('select-sub-entity', (detail) => (this.commit.selectedEntity = detail));
        this.onRemoveSubEntity = this.createListener('remove-sub-entity', (detail) => {
            if (this.entities === undefined || this.entities.length === 0) {
                return;
            }
            this.commit.entities = this.entities.filter(entity => entity !== detail);
        });
        /**
         * Some serialised store state.
         */
        this.toEntity = () => ({
            label: this.x,
            quantity: this.derived,
        });
        /**
         * Listeners not on the public interface can be created anonymously.
         */
        this.createAsyncListener('sub-load', (type) => __awaiter(this, void 0, void 0, function* () {
            return this.invokeAsyncAPI(() => __awaiter(this, void 0, void 0, function* () {
                this.commit.lastChangedProperty = type;
                return this.service.list();
            }), 'sub-load-error', 'sub-load-success');
        }));
        this.createListener('sub-load-success', entities => this.commit.entities = entities);
        this.createErrorListener('sub-load-error', () => this.commit.entities = undefined);
        /**
         * Stores support creating multiple listeners for the same key, so we can piggyback on events to drive other logic.
         */
        this.createListener([
            'y-changed',
            'z-changed',
        ], (_, { type }) => this.emit('sub-load', type));
    }
    /**
     * Does not need any special code as it's computing based on properties that are already observable.
     */
    get derived() {
        return this.y + this.z;
    }
};
__decorate([
    observable
], DefaultSubStore.prototype, "x", void 0);
__decorate([
    observable
], DefaultSubStore.prototype, "y", void 0);
__decorate([
    observable
], DefaultSubStore.prototype, "z", void 0);
__decorate([
    observable
], DefaultSubStore.prototype, "lastChangedProperty", void 0);
__decorate([
    observable
], DefaultSubStore.prototype, "entities", void 0);
__decorate([
    observable
], DefaultSubStore.prototype, "selectedEntity", void 0);
DefaultSubStore = __decorate([
    __param(0, SubService)
], DefaultSubStore);
export { DefaultSubStore };
/**
 * Registering as `transient` with the DI directly instead of via `registerStore` given test suite create/destroy cycle.
 * Normally store fragments are singletons, use `registerStore` in 99% of cases.
 * @internal
 */
export const SubStore = DI.createInterface(x => x.transient(DefaultSubStore));
/**
 * @internal
 */
let DefaultTestStore = class DefaultTestStore extends AbstractStoreRoot {
    constructor(subStore, service) {
        /**
         * Connect the sub store node to the tree by passing it.
         */
        super(subStore);
        this.subStore = subStore;
        this.service = service;
        this.a = '';
        this.b = 0;
        this.c = 0;
        this.onAChanged = this.createListener('a-changed', (detail) => (this.commit.a = detail));
        this.onBChanged = this.createListener('b-changed', (detail) => (this.commit.b = detail));
        this.onCChanged = this.createListener('c-changed', (detail) => (this.commit.c = detail));
        /**
         * Listeners not on the public interface can be created anonymously.
         */
        this.createAsyncListener('post-report', (detail) => __awaiter(this, void 0, void 0, function* () {
            yield this.service.send(detail);
        }));
        /**
         * Stores support creating multiple listeners for the same key, so we can piggyback on events to drive other logic.
         */
        this.createListener([
            'a-changed',
            'b-changed',
        ], (_, { type, detail }) => this.emit('post-report', { trigger: type, value: detail }));
    }
    /**
     * Marked as volatile as it contains branching code paths
     */
    get volatileDerived() {
        return this.a === 'multiply' ? this.b * this.c : this.b + this.c;
    }
    /**
     * Generates css classnames based on store state
     */
    get classNames() {
        return classNames(['warning', this.volatileDerived > 100]);
    }
};
__decorate([
    observable
], DefaultTestStore.prototype, "a", void 0);
__decorate([
    observable
], DefaultTestStore.prototype, "b", void 0);
__decorate([
    observable
], DefaultTestStore.prototype, "c", void 0);
__decorate([
    volatile
], DefaultTestStore.prototype, "volatileDerived", null);
DefaultTestStore = __decorate([
    __param(0, SubStore),
    __param(1, ReportService)
], DefaultTestStore);
export { DefaultTestStore };
/**
 * Registering as `transient` with the DI directly instead of via `registerStore` given test suite create/destroy cycle.
 * Normally store fragments are singletons, use `registerStore` in 99% of cases.
 * @internal
 */
export const TestStore = DI.createInterface(x => x.transient(DefaultTestStore));
