import { __awaiter, __decorate } from "tslib";
import { createErrorMap, isDev } from '@genesislcap/foundation-utils';
import { defaultExecutionContext, observable, Observable as FASTObservable, } from '@microsoft/fast-element';
import { DI } from '@microsoft/fast-foundation';
import safeStringify from 'fast-safe-stringify';
import { of } from 'rxjs';
import { distinctUntilChanged, map } from 'rxjs/operators';
import { logger } from '../utils';
const getBindArgs = (...args) => {
    const [token] = args;
    let binding = (x) => x;
    if (typeof token === 'function') {
        binding = token;
    }
    else if (typeof token === 'string') {
        binding = (x) => x[token];
    }
    return binding;
};
/**
 * @internal
 */
export var EventListenerType;
(function (EventListenerType) {
    EventListenerType["sync"] = "sync";
    EventListenerType["async"] = "async";
    EventListenerType["error"] = "error";
})(EventListenerType || (EventListenerType = {}));
/**
 * Creates a dependency injection key for the store being registered.
 * @param Base - The store fragment class.
 * @param name - The name of your store fragment, mostly used for internal logging.
 * @returns The created key.
 *
 * @example
 * ```ts
 * export const TradeEntry = registerStore<TradeEntry>(DefaultTradeEntry, 'TradeEntry');
 * ```
 *
 * @public
 */
export const registerStore = (Base, name = 'store') => {
    class Store extends Base {
        get name() {
            return name;
        }
    }
    return DI.createInterface((x) => x.singleton(Store));
};
const logValue = (value) => safeStringify(value, undefined, 2, {
    depthLimit: 1,
    edgesLimit: 1,
});
const internalEmitOptions = {
    bubbles: true,
    composed: true,
    cancelable: true,
};
/**
 * The abstract store that concrete store fragments must extend, which differs from the {@link AbstractStoreRoot}.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 * @typeParam TStoreRoot - The store root interface.
 *
 * @public
 */
export class AbstractStore {
    /**
     * @param storeFragments - The child store fragments this store fragment will manage.
     */
    constructor(...storeFragments) {
        /**
         * The store fragment's event listener map.
         *
         * @remarks
         * Keyed by listener to allow events to have multiple listeners for cleaner logic flows.
         * See multiple event keys example in `createAsyncListener`.
         *
         * @internal
         */
        this.eventListenerMap = new Map();
        /**
         * The child store fragments this store fragment manages.
         *
         * @internal
         */
        this.storeFragments = [];
        /**
         * The value commit proxy handler.
         *
         * @internal
         * @privateRemarks
         * This provides a way to do things on value setting like logging, validation, reject or tweak values, have a
         * middleware type setup to perhaps add dev tool state time travel functionality etc. Create a hash of current store
         * state and push to history for example. If the dev fails to use this commit proxy, ie. this.commit.prop = value,
         * it's not the end of the world as value will still be set. this.commit is typed the same as the main interface.
         * Commit mutations should always be synchronous.
         */
        this.commitHandler = {
            set: (target, prop, next) => {
                const prev = target[prop];
                if (next === prev) {
                    return true;
                }
                if (isDev()) {
                    logger.debug(`\n'${this.name}':\n   Committed ${prop} = ${logValue(next)}.\n   Previous ${prop} = ${logValue(prev)}`);
                }
                target[prop] = next;
                return true;
            },
        };
        /**
         * The value commit proxy.
         *
         * @example
         * ```ts
         * this.commit.propX = value;
         * ```
         *
         * @beta
         * @remarks
         * this.commit has the same interface as the store itself, so props are strongly typed.
         */
        this.commit = new Proxy(this, this.commitHandler);
        /** {@inheritDoc Store.errors} */
        this.errors = createErrorMap(logger.error);
        /**
         * Creates an event listener by type.
         *
         * @typeParam TDetail - The CustomEvent detail.
         * @param type - The type of event listener to create. See {@link EventListenerType}.
         * @param keys - The event key or keys from the store fragment's event detail map.
         * @param token - The function handling the event.
         * @returns The event listener when one key is passed or undefined.
         *
         * @internal
         */
        this.createListenerType = (type, keys, token) => {
            let listener;
            const listenerKeys = Array.isArray(keys) ? keys : [keys];
            listenerKeys.forEach((key) => {
                listener = this.createListenerEnvelope(type, token);
                this.eventListenerMap.set(listener, key);
            });
            return listenerKeys.length === 1 ? listener : undefined;
        };
        /**
         * Creates an event listener envelope.
         *
         * @remarks
         * These wrap the provided token.
         *
         * @typeParam TDetail - The CustomEvent detail.
         * @param listenerType - The type of event listener to create. See {@link EventListenerType}.
         * @param token - The function handling the event.
         * @returns An event listener.
         *
         * @internal
         */
        this.createListenerEnvelope = (listenerType, token) => {
            if (listenerType === EventListenerType.async) {
                return (event) => __awaiter(this, void 0, void 0, function* () {
                    const { detail, type } = event;
                    this.logEvent(type, true);
                    return token(detail, event);
                });
            }
            return (event) => {
                const { detail, type } = event;
                this.logEvent(type);
                if (listenerType === EventListenerType.error) {
                    this.errors.set(type, detail);
                }
                return token && token(detail, event);
            };
        };
        /**
         * Creates an event listener.
         *
         * @remarks
         * See the `createAsyncListener` for usage examples.
         *
         * @typeParam TDetail - The CustomEvent detail.
         * @param keys - The event key or keys from the store fragment's event detail map.
         * @param token - The function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * You can think of this like a `reducer` in the redux sense. You are allowed to commit values to the store in these
         * synchronous handlers.
         */
        this.createListener = (keys, token) => this.createListenerType(EventListenerType.sync, keys, token);
        /**
         * Creates an async event listener.
         *
         * @example
         * Creating an interface defined handler for a single event key.
         * ```ts
         * onDomainAction = this.createAsyncListener<DomainActionDetail>(
         *   'domain-action',
         *   async ({ id, message }) =>
         *     this.invokeAsyncAPI(
         *       async () => this.domainService.action(id, message),
         *       'domain-action-error',
         *       'domain-action-success'
         *     )
         * );
         * ```
         *
         * @example
         * Creating an anonymous handler in the constructor for multiple event keys.
         * ```ts
         * this.createAsyncListener(
         *   [
         *     'columns-changed',
         *     'types-changed',
         *     'max-rows-changed',
         *     'max-view-changed',
         *     'order-by-changed',
         *     'reverse-changed',
         *   ],
         *   async (_, { type }) => this.emit('domain-load')
         * );
         * ```
         *
         * @typeParam TDetail - The CustomEvent detail.
         * @param keys - The event key or keys from the store fragment's event detail map.
         * @param token - The async function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * You can think of this like an `effect` in the redux sense. You should not commit values to the store in these,
         * instead raise subsequent events to be handled synchronously, where commits are allowed.
         */
        this.createAsyncListener = (keys, token) => this.createListenerType(EventListenerType.async, keys, token);
        /**
         * Creates an error event listener.
         * @typeParam TDetail - The CustomEvent detail.
         * @param keys - The event key or keys from the store fragment's event detail map.
         * @param token - The function handling the event.
         * @returns The event listener.
         *
         * @public
         * @remarks
         * This logs and stores errors by event key in the store fragment's {@link ErrorMap}, allowing multiple errors to
         * co-exist, and be presentable to the user via the UI for further action or dismissal.
         */
        this.createErrorListener = (keys, token) => this.createListenerType(EventListenerType.error, keys, token);
        this.storeFragments = [...this.storeFragments, ...storeFragments];
    }
    /** {@inheritDoc Store.addStoreFragments} */
    addStoreFragments(...storeFragments) {
        this.storeFragments = [...this.storeFragments, ...storeFragments];
        if (this.root.element) {
            this.connectStoreFragments(this.root, storeFragments);
        }
    }
    /** {@inheritDoc Store.removeStoreFragments} */
    removeStoreFragments(...storeFragments) {
        this.storeFragments = this.storeFragments.filter((store) => !storeFragments.includes(store));
        if (this.root.element) {
            this.disconnectStoreFragments(this.root, storeFragments);
        }
    }
    /**
     * Alternative value commit api.
     * @param key - The property key from the store's interface.
     * @param value - The value to set.
     *
     * @beta
     * @privateRemarks
     * We should have only one way to do things, so we can ga gather user feedback and keep or kill this.
     */
    commitValue(key, value) {
        this.commit[key] = value;
    }
    /** {@inheritDoc Store.binding} */
    binding(token, subscriberChangeCallback, isVolatileBinding = false, context = defaultExecutionContext) {
        const valueGetter = getBindArgs(token);
        const initialSubscriber = {
            handleChange: (binding, args) => {
                if (subscriberChangeCallback) {
                    subscriberChangeCallback(binding(this, context), args);
                }
            },
        };
        const bindingObserver = FASTObservable.binding(valueGetter, initialSubscriber, isVolatileBinding);
        bindingObserver.observe(this, context);
        return bindingObserver;
    }
    /** {@inheritDoc Store.bindingAsRx} */
    bindingAsRx(token) {
        const valueGetter = getBindArgs(token);
        return of(this).pipe(map((store) => valueGetter(store)), distinctUntilChanged());
    }
    /** {@inheritDoc StoreConnectable.connect} */
    connect(root) {
        this.root = root;
        this.addEventListeners(root.element);
        this.connectStoreFragments(root);
    }
    /** {@inheritDoc StoreConnectable.disconnect} */
    disconnect(root) {
        this.removeEventListeners(root.element);
        this.disconnectStoreFragments(root);
        this.root = undefined;
    }
    /** {@inheritDoc Store.name} */
    get name() {
        return 'store';
    }
    /**
     * Connects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     * @param storeFragments - Optional store fragments. Defaults to this.storeFragments. Used for lazy add.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    connectStoreFragments(root, storeFragments = this.storeFragments) {
        storeFragments.forEach((store) => {
            const connectableStore = store;
            connectableStore.connect(root);
        });
    }
    /**
     * Disconnects the store fragment's child store fragments if any.
     * @param root - The store root fragment.
     * @param storeFragments - Optional store fragments. Defaults to this.storeFragments. Used for lazy remove.
     *
     * @internal
     * @privateRemarks
     * We have limited the connect and disconnect apis to store root only, so we need to cast these to invoke the same.
     */
    disconnectStoreFragments(root, storeFragments = this.storeFragments) {
        storeFragments.forEach((store) => {
            const connectableStore = store;
            connectableStore.disconnect(root);
        });
    }
    /**
     * Logs events.
     * @param key - The event key from the store fragment's event detail map.
     * @param isAsyncHandler - A flag to indicate if the handler is async.
     *
     * @internal
     */
    logEvent(key, isAsyncHandler) {
        if (isDev()) {
            logger.debug(`'${this.name}': Handle '${String(key)}' store event ${isAsyncHandler ? '[async]' : ''}`);
        }
    }
    /**
     * Add the store fragment's event listeners to the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via connect() when the root element changes.
     */
    addEventListeners(element) {
        for (const [listener, key] of this.eventListenerMap) {
            element.addEventListener(key, listener);
        }
    }
    /**
     * Removes the store fragment's event listeners from the root element.
     * @param element - The store root element.
     *
     * @internal
     * @privateRemarks
     * Called via disconnect() when the root element changes.
     */
    removeEventListeners(element) {
        for (const [listener, key] of this.eventListenerMap) {
            element.removeEventListener(key, listener);
        }
    }
    /**
     * Emit events to the stores directly via the standard event flow.
     * @param key - The event key from the store fragment's event detail map.
     * @param detail - The event detail object to send with the event.
     * @param options - The event options. By default bubbles and composed.
     *
     * @public
     * @remarks
     * By default we allow stores to emit everything in their TEventDetailMap, however TInternalEventDetailMap can be
     * provided to define additional internal events like x-success, x-error etc, or public events from other store
     * fragments.
     */
    emit(...args) {
        const [key, detail] = args;
        const hasDetail = detail !== undefined;
        if (isDev()) {
            logger.debug(`'${this.name}': Emit '${key}' store event${hasDetail ? ` with ${logValue(detail)}` : ''}`);
        }
        this.root.element.dispatchEvent(new CustomEvent(key, Object.assign({ detail }, internalEmitOptions)));
    }
    /**
     * A convenience method to invoke an async api and emit success and error events.
     *
     * @remarks
     * The async api function should throw when it encounters an error.
     *
     * @example
     * ```ts
     * onLoad = this.createAsyncListener('domain-load', async () =>
     *  this.invokeAsyncAPI(
     *    async () => {
     *      !this.domainService.initialized && (await this.domainService.initialize(this.asDomainServiceInit()));
     *      return this.domainService.list();
     *    },
     *    'domain-load-error',
     *    'domain-load-success'
     *  )
     * );
     * ```
     *
     * @param api - The async service api function.
     * @param error - The event key from the store fragment's event detail map.
     * @param success - The event key from the store fragment's event detail map.
     * @public
     */
    invokeAsyncAPI(api, error, success) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const result = yield api();
                // @ts-ignore (we know more about this than ts)
                success && this.emit(success, result);
            }
            catch (e) {
                // @ts-ignore (we know more about this than ts)
                this.emit(error, e);
            }
        });
    }
}
/**
 * The abstract store root that concrete store roots must extend.
 * @typeParam TStore - The store interface.
 * @typeParam TEventDetailMap - The store's public event detail map.
 * @typeParam TInternalEventDetailMap - The store's internal event detail map.
 *
 * @public
 */
export class AbstractStoreRoot extends AbstractStore {
    constructor() {
        super(...arguments);
        /** {@inheritDoc StoreRoot.ready} */
        this.ready = false;
        /**
         * @public
         * @privateRemarks
         * Set to self
         */
        this.root = this;
        this.onConnected = this.createListener('store-connected', (element) => (this.commit.element = element));
        this.onDisconnected = this.createListener('store-disconnected', () => (this.commit.element = undefined));
        this.onReady = this.createListener('store-ready', (ready) => (this.commit.ready = ready));
    }
    /**
     * Internal change handler for the element property.
     * @param prev - The previous element value.
     * @param next - The next element value.
     *
     * @internal
     * @privateRemarks
     * Manages the store listeners on change.
     */
    elementChanged(prev, next) {
        if (prev) {
            this.removeEventListeners(prev);
            this.disconnectStoreFragments(this);
        }
        if (next) {
            this.addEventListeners(next);
            this.connectStoreFragments(this);
        }
    }
}
__decorate([
    observable
], AbstractStoreRoot.prototype, "ready", void 0);
__decorate([
    observable
], AbstractStoreRoot.prototype, "element", void 0);
